package lowlevelfunc.boolfun;

/**
 * Class which represents a generic boolean function, along with its
 * representations (truth table, ANF), cryptographic properties (balancedness,
 * algebraic degree, nonlinearity, resiliency, propagation criterion, number of
 * nonzero linear structures), chaos properties (permutivity) and performances
 * as a CA-based PRNG rule over the ENT tests. The parameters of the PRNG are
 * supposed to be standard:
 * 
 * 1) The CA is composed by 64 cells, with periodic boundary conditions.
 * 2) The ENT tests are run on a single sequence produced as follows: 
 *    2.1) The CA is initialized with an initial configuration containing
 *         only a 1 in the 32nd cell.
 *    2.2) The CA is evolved for 65536 iterations.
 *    2.3) The trace of the 32nd cell is taken as the pseudorandom sequence.
 * 
 * @author Luca Mariot
 * @version 1.0
 * 
 */

import java.math.BigInteger;

public class BooleanFunction {
    
    //Representation-related attributes.
    private BigInteger deccode;
    private int nvar;
    private int ninputs;
    private boolean[] ttable;
    private boolean[] anfcoeffs;
    private String anfexpr;
    
    //Transforms-related attributes.
    private int[] poltable;
    private int[] whcoeffs;
    private int sprad;
    private int nlin;
    private int[] accoeffs;
    private int acmax;
    private int ssi; //Sum of square indicator
    
    //Crypto-related attributes.
    private boolean isBalanced;
    private int algdeg;
    private int nlinstruct;
    private int[] cid;
    private int[] pcd;
    
    //CA related attributes (these refer to the length of the CA and the
    //pseudorandom sequence generated for the ENT tests results).
    private int cacells;
    private int seqlength;
    
    //ENT tests attributes.
    private double entropy;
    private double chisqsum;
    private double chisqprob;
    private double meandev;
    private double pidev;
    private double scc;
    
    //Heuristic methods attributes.
    private double fitness;
    
    /** 
     * First constructor, decimal representation-based.
     * 
     * @param deccode   decimal representation code of the function.
     * @param nvar      number of variables of the function.
     */
    public BooleanFunction(BigInteger deccode, int nvar) {
        
        this.deccode = deccode;
        this.nvar = nvar;
        
        ninputs = (int)Math.pow(2,nvar);
        ttable = BinTools.dec2Bin(deccode, ninputs);
        poltable = BinTools.bin2Pol(ttable);
        
        whcoeffs = new int[ninputs];
        accoeffs = new int[ninputs];
        anfcoeffs = new boolean[ninputs];
        cid = new int[nvar];
        pcd = new int[nvar];
        
    }
    
    /** 
     * Second constructor, truth table representation-based.
     * 
     * @param ttable    truth table of the function (LSBF order)
     * @param nvar      number of variables of the function.
     */
    public BooleanFunction(boolean[] ttable, int nvar) {
        
        this.ttable = ttable;
        this.nvar = nvar;
        
        ninputs = ttable.length;
        deccode = BinTools.bin2DecBig(ttable);
        poltable = BinTools.bin2Pol(ttable);
        
        whcoeffs = new int[ninputs];
        accoeffs = new int[ninputs];
        anfcoeffs = new boolean[ninputs];
        cid = new int[nvar];
        pcd = new int[nvar];
        
    }
    
    //Getters.

    public int[] getAccoeffs() {
        return accoeffs;
    }

    public int getAcmax() {
        return acmax;
    }

    public int getAlgdeg() {
        return algdeg;
    }

    public boolean[] getAnfcoeffs() {
        return anfcoeffs;
    }

    public String getAnfexpr() {
        return anfexpr;
    }

    public double getChisqprob() {
        return chisqprob;
    }

    public double getChisqsum() {
        return chisqsum;
    }

    public BigInteger getDeccode() {
        return deccode;
    }

    public double getEntropy() {
        return entropy;
    }

    public boolean isIsBalanced() {
        return isBalanced;
    }

    public double getMeandev() {
        return meandev;
    }

    public int getNinputs() {
        return ninputs;
    }

    public int getNlinstruct() {
        return nlinstruct;
    }

    public int getNvar() {
        return nvar;
    }

    public double getPidev() {
        return pidev;
    }

    public int[] getPoltable() {
        return poltable;
    }

    public double getScc() {
        return scc;
    }

    public int getSprad() {
        return sprad;
    }

    public boolean[] getTtable() {
        return ttable;
    }

    public int[] getWhcoeffs() {
        return whcoeffs;
    }

    public int getNlin() {
        return nlin;
    }

    public int[] getCid() {
        return cid;
    }

    public int[] getPcd() {
        return pcd;
    }

    public int getCacells() {
        return cacells;
    }

    public int getSeqlength() {
        return seqlength;
    }

    public double getFitness() {
        return fitness;
    }
    
    //Setters.

    public void setAccoeffs(int[] accoeffs) {
        this.accoeffs = accoeffs;
    }

    public void setAcmax(int acmax) {
        this.acmax = acmax;
    }

    public void setAlgdeg(int algdeg) {
        this.algdeg = algdeg;
    }

    public void setAnfcoeffs(boolean[] anfcoeffs) {
        this.anfcoeffs = anfcoeffs;
    }

    public void setAnfexpr(String anfexpr) {
        this.anfexpr = anfexpr;
    }

    public void setChisqprob(double chisqprob) {
        this.chisqprob = chisqprob;
    }

    public void setChisqsum(double chisqsum) {
        this.chisqsum = chisqsum;
    }

    public void setDeccode(BigInteger deccode) {
        this.deccode = deccode;
    }

    public void setEntropy(double entropy) {
        this.entropy = entropy;
    }

    public void setIsBalanced(boolean isBalanced) {
        this.isBalanced = isBalanced;
    }

    public void setMeandev(double meandev) {
        this.meandev = meandev;
    }

    public void setNinputs(int ninputs) {
        this.ninputs = ninputs;
    }

    public void setNlinstruct(int nlinstruct) {
        this.nlinstruct = nlinstruct;
    }

    public void setNvar(int nvar) {
        this.nvar = nvar;
    }

    public void setPidev(double pidev) {
        this.pidev = pidev;
    }

    public void setPoltable(int[] poltable) {
        this.poltable = poltable;
    }

    public void setScc(double scc) {
        this.scc = scc;
    }

    public void setSprad(int sprad) {
        this.sprad = sprad;
    }

    public void setTtable(boolean[] ttable) {
        this.ttable = ttable;
    }

    public void setWhcoeffs(int[] whcoeffs) {
        this.whcoeffs = whcoeffs;
    }

    public void setNlin(int nlin) {
        this.nlin = nlin;
    }

    public void setCid(int[] cid) {
        this.cid = cid;
    }

    public void setPcd(int[] pcd) {
        this.pcd = pcd;
    }

    public void setCacells(int cacells) {
        this.cacells = cacells;
    }

    public void setSeqlength(int seqlength) {
        this.seqlength = seqlength;
    }

    public void setFitness(double fitness) {
        this.fitness = fitness;
    }

    public int getSsi() {
        return ssi;
    }

    public void setSsi(int ssi) {
        this.ssi = ssi;
    }
    
}
