package softcomp.genalg;

/**
 * Class containing various methods for combination-based GA.
 * 
 * @author Luca Mariot
 */

import lowlevelfunc.boolfun.*;
import java.util.Vector;
import java.util.Random;

public class CombGATools {
    
    /**
     * Build the profile (n,m,d,nl,r,sprad) of a plateaued boolean function starting
     * from the number of input variables n.
     * 
     * @param nvar number of input variables
     * @return an array of 6 int components specifiying number of variables n,
     *         resiliency order m, algebraic degree d, nonlinearity nl, plateuau
     *         parameter r and spectral radius sprad
     */
    public static int[] buildPlatProfile(int nvar) {
        
        int[] profile = new int[6];
        profile[0] = nvar;      //First component: number of variables
   
        int r=0;    //parameter for plateaued functions profile
        
        //If nvar is even, then set r to nvar/2 + 1 (otherwise we get bent
        //functions). Else, set r to (nvar+1)/2
        if((nvar % 2) == 0) {
            r = (nvar/2) + 1;  //  
        } else {                         //--> 5th component: plateau parameter
            r = (nvar+1) / 2; //
        }
        int sprad = (int)Math.pow(2,r);
        int m = r - 2;
        int d = nvar - m - 1;
        int nl = (int)Math.pow(2,nvar-1) - sprad/2;
        
        profile[1] =  m;    //2nd component: resiliency order m
        profile[2] = d;  //3rd component: algebraic degree
        profile[3] = nl; //4th component: nonlinearity
        profile[4] = r;  //5th component: parameter r
        profile[5] = sprad; //6th component: spectral radius sprad
        
        return profile;
        
    }
    
    /**
     * Count the number of positions set to zero in the Walsh spectrum to satisfy
     * resiliency order m.
     * 
     * @param nvar  number of input variables of the boolean function
     * @param m     resiliency order of the boolean function
     * @return the number of positions in the Walsh spectrum to be set to zero
     *         in order to meet res(m)
     */
    public static int countResZeros(int nvar, int m) {
        
        int reszeros = 1;
        for(int i=1; i<=m; i++) {
            
            reszeros += CombTools.binCoeff(nvar, i);
            
        }
        
        return reszeros;
        
    }
    
    /**
     * Return a vector containing the position in the permutable positions in
     * the Walsh spectrum. These positions are the ones having Hamming weight
     * greater than m, where m is the resiliency order.
     * 
     * @param nvar  number of input variables of the boolean function
     * @param m     resiliency order of the boolean function
     * @return a Vector of Integer containing the ordered sequence of permutable
     *         positions in the Walsh spectrum
     */
    public static int[] findPermutablePos(int nvar, int m, int rstlen) {
        
        int[] rstpos = new int[rstlen];
        
        //Walsh spectrum length
        int speclen = (int)Math.pow(2,nvar);
        
        int j=0;
        
        for(int i=0; i<speclen; i++) {
            
            //Convert the current index in binary
            boolean[] binind = BinTools.dec2BinMod(i, nvar);
            
            //If the hamming weight of the index is higher than the resiliency
            //order, add it to the vector
            if((BinTools.hwt(binind) > m) && (j < rstlen)) {
                rstpos[j] = i;
                j++;
            }
            
        }
        
        //Convert the array
        return rstpos;
        
    }
    
    /** Count the number of the remaining positions to be set respectively to
     * zero, -Wmax and +Wmax in the Walsh spectrum
     * 
     * @param nvar          number of input variables of the boolean function
     * @param sprad         spectral radius (Wmax) of the Walsh spectrum
     * @param speclen       Walsh spectrum length
     * @param rstspeclen    number of permutable positions in the spectrum
     * @return an int array specifying the number of positions to be set
     *         respectively to zero, -Wmax and +Wmax in the spectrum
     */
    public static int[] countWHCoeffsProp(int nvar, int sprad, int speclen,
            int rstspeclen) {
        
        int parsevalsum = (int)Math.pow(2.0, (2*nvar));     //sum yielded by Parseval's relation
        //System.out.println("Parseval = "+parsevalsum);
        int propnz = parsevalsum / (sprad*sprad);   //number of positions equal to +-Wmax
        //System.out.println("Nonzero positions = "+propnz);
        int propremz = rstspeclen - propnz; //number of zero positions for the remaining zeros
        int propnzneg = (propnz - (speclen/sprad)) / 2; //number of positions equal to -Wmax
        int propnzpos = propnzneg + (speclen/sprad);    //number of positions equal to +Wmax
        
        int[] toRet = {propremz, propnzneg, propnzpos};
        
        return toRet;
        
    }
    
    /**
     * Build the ordered spectrum.
     * 
     * @param propremz
     * @param propnzneg
     * @param propnzpos
     * @param sprad
     * @return 
     */
    public static int[] buildOrdSpec(int propremz, int propnzneg,
            int propnzpos, int sprad) {
        
        //Step 1: create the ordered spectrum vector
        Vector<Integer> ordspec = new Vector<Integer>();
        
        //Fill the remaining zeros
        for(int i=0; i<propremz; i++) {
            
            ordspec.add(0);
            
        }
        
        ordspec.trimToSize();
        
        //Fill the positions equal to -Wmax
        for(int i=0; i<propnzneg; i++) {
            
            ordspec.add(-sprad);
            
        }
        
        ordspec.trimToSize();
        
        //Fill the positions equal to +Wmax
        for(int i=0; i<propnzpos; i++) {
            
            ordspec.add(sprad);
            
        }
        
        ordspec.trimToSize();
        
        int[] ordarr = new int[ordspec.capacity()];
        
        for(int i=0; i<ordarr.length; i++) {
            
            ordarr[i] = ordspec.elementAt(i);
            
        }
        
        return ordarr;
        
    }
    
    /**
     * Return the arrays containing the positions of 0s, -sprad and +sprad in
     * a restricted Walsh spectrum
     * 
     * @param rstspec
     * @param propz
     * @param propn
     * @param propp
     * @param sprad
     * @return 
     */
    public static int[][] buildMapsCoeff(int[] rstspec, int propz, int propn,
            int propp, int sprad) {
        
        int[][] positions = new int[3][];
        positions[0] = new int[propz];
        positions[1] = new int[propn];
        positions[2] = new int[propp];
        
        //Counters for indexing the three positions vectors
        int j=0;
        int k=0;
        int l=0;
        
        //Scan the restricted spectrum
        for(int i=0; i<rstspec.length; i++) {
            
            if((rstspec[i] == 0) && (j < propz)) {
                
                positions[0][j] = i;
                j++;
                
            } else {
                
                if((rstspec[i] == -sprad) && (k < propn)) {
                    
                    positions[1][k] = i;
                    k++;
                    
                } else {
                    
                    positions[2][l] = i;
                    l++;
                    
                }
                
            }
            
        }
        
        return positions;
        
    }
    
    /**
     * Create an individual for the permutation GA as a random permutation of
     * the restricted Walsh spectrum.
     * 
     * @param propremz  proportions of zeros in the restricted spectrum
     * @param propnzneg proportions of -Wmax in the restricted spectrum
     * @param propnzpos proportions of +Wmax in the restricted spectrum
     * @param sprad     spectral radius (Wmax) of the Walsh spectrum
     * @param genrand   random number generator
     * @return an int array specifying a random permutation of the restricted
     *         spectrum of a pseudoboolean plateaued function
     */
    public static int[] createRandomCombIndiv(int rstspeclen, int propremz,
            int propnzneg, int propnzpos, int sprad, Random genrand) {
        
        //Step 1: create the candidate positions vector
        Vector<Integer> candpos = new Vector<Integer>();
        
        for(int i=0; i<rstspeclen; i++) {
            
            candpos.add(i);
            
        }
        
        candpos.trimToSize();
        
        //Step 2: Assign value -sprad to propnzneg randomly chosen positions
        int[] rstspec = new int[rstspeclen];
        
        for(int i=0; i<propnzneg; i++) {
            
            candpos.trimToSize();
            
            //Randomly draw a position from the ordered vector
            int randpos = genrand.nextInt(candpos.capacity());
            
            //Set the drawn position to -sprad 
            rstspec[candpos.elementAt(randpos)] = -sprad;
            
            //System.out.print("Set position "+candpos.elementAt(randpos)+" to "+rstspec[candpos.elementAt(randpos)]+"; remaining positions: ");
            
            //Delete the position from the candidate vector
            candpos.removeElementAt(randpos);
            candpos.trimToSize();
            /*for(int j=0; j<candpos.capacity(); j++) {
                System.out.print(candpos.elementAt(j)+" ");
            }
            System.out.println("");*/
            
        }
        
        //Step 3: Assign value +sprad to propnzpos randomly chosen positions
        
        for(int i=0; i<propnzpos; i++) {
            
            candpos.trimToSize();
            
            //Randomly draw a position from the ordered vector
            int randpos = genrand.nextInt(candpos.capacity());
            
            //Set the drawn position to -sprad 
            rstspec[candpos.elementAt(randpos)] = sprad;
            
            //System.out.print("Set position "+candpos.elementAt(randpos)+" to "+rstspec[candpos.elementAt(randpos)]+"; remaining positions: ");
            
            //Delete the position from the candidate vector
            candpos.removeElementAt(randpos);
            candpos.trimToSize();
            /*for(int j=0; j<candpos.capacity(); j++) {
                System.out.print(candpos.elementAt(j)+" ");
            }
            System.out.println("");*/
            
        }
        
        //No further sampling is needed, since the remaining positions are
        //already initialised to zero :-)
        
        return rstspec;
        
    }
    
    /**
     * Create a random population of restricted Walsh spectra.
     * 
     * @param popsize   Size of the population
     * @param propremz  proportions of zeros in the restricted spectrum
     * @param propnzneg proportions of -Wmax in the restricted spectrum
     * @param propnzpos proportions of +Wmax in the restricted spectrum
     * @param sprad     spectral radius (Wmax) of the Walsh spectrum
     * @param genrand   random number generator
     * @return an array of WalshPermutation objects representing the population
     */
    public static WalshPermutation[] createWalshPopComb(int popsize, int propremz,
            int propnzneg, int propnzpos, int sprad, int rstspeclen, Random genrand) {
        
        WalshPermutation[] population = new WalshPermutation[popsize];
        
        for(int i=0; i<popsize; i++) {
            
            int[] rstspec = createRandomCombIndiv(rstspeclen, 
                    propremz, propnzneg, propnzpos, sprad, genrand);
            population[i] = new WalshPermutation(rstspec);
            
        }
        
        return population;
        
    }
    
    /**
     * Build the pseudo-boolean function starting from its restricted spectrum,
     * using the Inverse Walsh Transform
     * @param speclen   length of the spectrum
     * @param ordspec   ordered spectrum
     * @param nzpos     Vector of permutable positions in the spectrum
     * @param rstspec   restricted spectrum
     * @return a matrix containing the truth table of the pseudo-boolean
     *         function and its Walsh spectrum
     */
    public static int[][] decodeSpectrumComb(int speclen, int[] rstspecpos,
            int[] rstspec) {
        
        //Step 1: create the complete spectrum
        int[] func = new int[speclen];
        int[] spectrum = new int[speclen];
        
        for(int i=0; i<rstspecpos.length; i++) {
            
            func[rstspecpos[i]] = rstspec[i];
            spectrum[rstspecpos[i]] = rstspec[i];
            
        }
        
        //Step 2: apply the IWT and return the function
        BoolTransf.calcInvFWT(func, 0, speclen);
        
        int[][] toRet = {func,spectrum};
        
        return toRet;
        
    }
    
    /** 
     * Collapse a pseudo-boolean function to the nearest boolean function
     * 
     * @param pseudobf  truth table of a pseudo-boolean function
     * @return the truth table of the nearest boolean function
     */
    public static boolean[] collapsePseudoBoolFunc(int[] pseudobf) {
        
        boolean[] nearestbf = new boolean[pseudobf.length];
        
        for(int i=0; i<pseudobf.length; i++) {
            
            //Compute the value of the nearest bit for the current position
            boolean b = false;
            
            if(pseudobf[i] > 0) {
                
                b = false;                      //if the value of the current 
                                                //position is higher than 0,
                                                //then collapse it to +1 (0)
            } else {
                
                if(pseudobf[i] < 0) {
                    b = true;                   //if the value of the current 
                                                //position is lower than 0,
                                                //then collapse it to -1
                } else {
                    
                    //If the value of the current position equals 0,
                    //select a random value in {-1,+1}
                    b = Math.random() > 0.5;
                    
                }
            }
            
            nearestbf[i] = b;
            
        }
        
        return nearestbf;
        
    }
    
    public static int[][] roulWheelSelectionComb(WalshPermutation[] population,
            Random genrand) {
        
        //Compute the sum of fitnesses
        double sumfit = 0.0;
        
        for(int i=0; i<population.length; i++) {
            
            sumfit += population[i].getFitness();
            
        }
        
        int[][] candidates = new int[population.length][population[0].getRstspec().length];
        
        for(int i=0; i<candidates.length; i++) {
            
            //Select a point in the wheel.
            double landpt = genrand.nextDouble() * sumfit;
            
            //Find which slot of the wheel the landing point belongs to.
            int j = 0;
            double partsum = 0.0;
            
            while((partsum < landpt) && (j != population.length-1)) {
                j++;
                partsum += population[j].getFitness();
            }
            
            //index j now holds the position of the individual to be selected,
            //copy it in the candidates vector
            System.arraycopy(population[j].getRstspec(), 0, candidates[i], 0, candidates[i].length);
            
        }
        
        return candidates;
        
    }
    
    /**
     * Perform (deterministic) tournament selection on a population of
     * restricted spectra.
     * 
     * @param population    population of restricted Walsh spectra
     * @param tsize         tournament size
     * @param genrand       random numbers generator
     * @return an array of candidate restricted spectra for reproduction 
     */
    public static int[][] tournamentSelectionComb(WalshPermutation[] population,
            int tsize, Random genrand) {
        
        int[][] candidates = new int[population.length][population[0].getRstspec().length];
        
        //Main cycle: perform  tournament selection until the candidates array
        //is complete
        for(int i=0; i<candidates.length; i++) {
            
            //Step 1: randomly draw tsize individuals from the population
            //(with replacement) and select the one with minimum fitness
            int winner = 0;
            
            for(int j=0; j<tsize; j++) {
                
                //Select a random individual from the population
                int sel = genrand.nextInt(population.length);
                
                //At the first iteration, set the winner to the first individual
                //extracted
                if(j==0) {
                    winner = sel;
                }
                
                //For the subsequent iterations compare the fitness of the
                //current extracted individual with that of the current winenr
                if((j > 0) && population[sel].getFitness() < population[winner].getFitness()) {
                    
                    winner = sel;
                    
                }
                
            }
            
            //Step 2: copy the winner in the candidates population
            System.arraycopy(population[i].getRstspec(), 0, candidates[i], 0,
                    population[i].getRstspec().length);
            
        }
        
        return candidates;
        
    }
    
    /**
     * Perform one step of the permutation-based GA
     * 
     * @param pop
     * @param genrand
     * @param pcross
     * @param pmut
     * @param tsize
     * @param selmet
     * @param propz
     * @param propn
     * @param propp
     * @param sprad
     * @return 
     */
    public static WalshPermutation[] createNextGenComb(WalshPermutation[] pop,
            Random genrand, double pcross, double pmut, int tsize, int selmet,
            int propz, int propn, int propp, int sprad) {
        
        //Create the array for the next generation
        WalshPermutation[] nextgen = new WalshPermutation[pop.length];
        
        int[][] candidates;
        candidates = new int[pop.length][];
        
        switch(selmet) {
            
            case 1:     candidates = tournamentSelectionComb(pop, tsize, genrand);
                break;
            case 2:     candidates = roulWheelSelectionComb(pop, genrand);
                break;
            default:    System.exit(1);
                break;
        }
        
        //Main cycle: for each couple (i,i+1), determines whether 
        //crossover+mutation or mutation only should be applied.
        for(int i = 0; i < pop.length; i+=2) {
            
            int[] parent1 = new int[candidates[i].length];
            int[] parent2 = new int[candidates[i+1].length];
            
            System.arraycopy(candidates[i], 0, parent1, 0, candidates[i].length);
            System.arraycopy(candidates[i+1], 0, parent2, 0, candidates[i+1].length); 
            
            if(genrand.nextDouble() < pcross) {
                
                //perform crossover AND mutation
                int[][] childs = new int[2][];
                
                childs[0] = CrossoverOperators.unifCrossoverSpectra(parent1,
                        parent2, propz, propn, propp, sprad);
                childs[1] = CrossoverOperators.unifCrossoverSpectra(parent1,
                        parent2, propz, propn, propp, sprad);
                
                //Build maps of 0s, -sprad and +sprad
                int[][] c0maps = buildMapsCoeff(childs[0], propz, propn, propp, sprad);
                int[][] c1maps = buildMapsCoeff(childs[1], propz, propn, propp, sprad);
                
                MutationOperators.switchMutationComb(childs[0], c0maps[0],
                        c0maps[1], c0maps[2], pmut, sprad, genrand);
                MutationOperators.switchMutationComb(childs[1], c1maps[0],
                        c1maps[1], c1maps[2], pmut, sprad, genrand);
                
                //Copy the two children in the next generation array.
                nextgen[i] = new WalshPermutation(childs[0]);
                nextgen[i+1] = new WalshPermutation(childs[1]);
                
            } else {
                
                //Perform only mutation of the parents.
                
                //Build maps of 0s, -sprad and +sprad
                int[][] p1maps = buildMapsCoeff(parent1, propz, propn, propp, sprad);
                int[][] p2maps = buildMapsCoeff(parent2, propz, propn, propp, sprad);
                
                MutationOperators.switchMutationComb(parent1, p1maps[0],
                        p1maps[1], p1maps[2], pmut, sprad, genrand);
                MutationOperators.switchMutationComb(parent2, p2maps[0],
                        p2maps[1], p2maps[2], pmut, sprad, genrand);
                
                //Copy the two mutated parents in the next generation array.
                nextgen[i] = new WalshPermutation(parent1);
                nextgen[i+1] = new WalshPermutation(parent2);
                
            }
            
        }
        
        return nextgen;
        
    }    
    
}
