package softcomp.genalg;

/**
 * Class containing various crossover methods, both for balanced functions and
 * bipermutive functions.
 * 
 * @author Luca Mariot
 * @version 1.0
 */

import java.math.BigInteger;
import java.util.Random;
import java.util.Vector;
import lowlevelfunc.boolfun.*;

public class CrossoverOperators {
    
    /**
     * Search for an int value in a int vector, returning its position (-1 is
     * returned if it is not contained).
     * 
     * @param vect  A vector of int
     * @param val   The int value to be searched.
     * @return      The position of value val in vect if it is present, -1
     *              otherwise.
     */
    public static int positionOf(int val, Vector<Integer> vect) {
        
        for(int i=0; i<vect.capacity(); i++) {
            
            if(vect.elementAt(i) == val) {
                return i;
            }
            
        }
        
        return -1;
        
    }
    
    /**
     * Find the position of a value in an array.
     * 
     * @param val value to be searched
     * @param arr array of integers
     * @return index of val in arr (-1 if val is not contained in arr)
     */
    public static int findPos(int val, int[] arr) {
    
        int i=0;
        
        while(i<arr.length) {
            
            if(arr[i] == val) {
                
                return i;
                
            }
            
            i++;
            
        }
        
        return -1;
        
    }
    
    /**
     * Randomly select a gene from two parents with uniform probability
     * 
     * @param par1  first parent
     * @param par2  second parent
     * @param i     position to be crossed
     * @return 
     */
    public static int randomCrossSingleGene(int[] par1, int[] par2, int i,
            Vector<Integer> difpos) {
        
        if(Math.random()<=0.5) {
            return par1[difpos.elementAt(i)];
        } else {
            return par2[difpos.elementAt(i)];
        }
        
    }
    
    /**
     * Cross a single gene between two parents when one of the three possible
     * values has reached the maximum number allowed.
     * 
     * @param par1
     * @param par2
     * @param i
     * @param val
     * @param sprad
     * @return 
     */
    public static int constr1CrossSingleGene(int[] par1, int[] par2, int i,
            int val, Vector<Integer> difpos) {
        
        //If the i-th gene of the first parent equals the value whose
        //threshold has been achieved, copy in the child the value of the
        //second parent
        if(par1[difpos.elementAt(i)] == val) {

            return par2[difpos.elementAt(i)];

        } else {

            //If the i-th gene of the second parent equals the value whose
            //threshold has been achieved, copy in the child the value of the
            //first parent
            if(par2[difpos.elementAt(i)] == val) {

                return par1[difpos.elementAt(i)];

            } else {

                //In all other cases, copy from one of the two parents by
                //randomly choosing them
                return randomCrossSingleGene(par1, par2, i, difpos);

            }

        }
        
    }
    
    /**
     * Perform crossover between two restricted Walsh spectrum permutations
     * 
     * @param par1
     * @param par2
     * @param propz
     * @param propn
     * @param propp
     * @return 
     */
    public static int[] unifCrossoverSpectra(int[] par1, int[] par2, int propz,
            int propn, int propp, int sprad) {
        
        int[] child = new int[par1.length];
        
        //Step 1: find common and different positions between the two parents,
        //and set the child equal to the parents in the common positions
        Vector<Integer> compos = new Vector<Integer>();
        Vector<Integer> difpos = new Vector<Integer>();
        int comz = 0;
        int comn = 0;
        int comp = 0;
        
        for(int i=0; i<par1.length; i++) {
            
            if(par1[i] == par2[i]) {
                
                compos.add(i);
                
                //Set the child equal to the parents in this common position
                child[i] = par1[i];
                
                //Update counters for common positions
                if(par1[i] == 0) {
                    comz++;
                } else if(par1[i] == -sprad) {
                    comn++;
                } else {
                    comp++;
                }
                
            } else {
                //Add the index to the different positions
                difpos.add(i);
            }
            
        }
        
        compos.trimToSize();
        difpos.trimToSize();
        
        //Compute the number of remaining 0s, -sprad and +sprad to be set
        int remz = propz - comz;
        int remn = propn - comn;
        int remp = propp - comp;
        
        //Step 2: apply crossover on the different positions
        int countz = 0;
        int countn = 0;
        int countp = 0;
        
        for(int i = 0; i<difpos.capacity(); i++) {
            
            //Encode the possible conditions on the counters in an integer
            boolean[] condbool = {countz==remz, countn==remn, countp==remp};
            int cond = BinTools.bin2Dec(condbool);
            //Main switch on the possible combinations of conditions
            switch(cond) {
                
                //Case 0: (countz<remz && countn<remn && countp<remp)
                //All counters are below the respective threshold.
                //Randomly choose the first or second parent and copy its i-th
                // gene in the child
                case 0: {
                    
                    child[difpos.elementAt(i)] = randomCrossSingleGene(par1,
                            par2, i, difpos);
                    
                    //Update counters
                    if(child[difpos.elementAt(i)] == 0) {
                        countz++;
                    } else {
                        if(child[difpos.elementAt(i)] == -sprad) {
                            countn++;
                        } else {
                            countp++;
                        }
                    }
                    
                } break;
                  
                //Case 1: (countz=remz && countn<remn && countp<remp)
                //Maximum number of 0s reached
                case 1: {
                    
                    child[difpos.elementAt(i)] = constr1CrossSingleGene(par1, 
                            par2, i, 0, difpos);
                    
                    //Update counters
                    if(child[difpos.elementAt(i)] == -sprad) {
                        countn++;
                    } else {
                        countp++;
                    }
                    
                } break;
                
                //Case 2: (countz<remz && countn=remn && countp<remp)
                //Maximum number of -sprad reached
                case 2: {
                    
                    child[difpos.elementAt(i)] = constr1CrossSingleGene(par1, 
                            par2, i, -sprad, difpos);
                    
                    //Update counters
                    if(child[difpos.elementAt(i)] == sprad) {
                        countp++;
                    } else {
                        countz++;
                    }
                    
                } break;
                
                //Case 3: (countz=remz && countn=remn && countp<remp)
                //Maximum number of 0s and -sprad reached
                case 3: {
                    
                    //Copy +sprad in the childen
                    child[difpos.elementAt(i)] = sprad;
                    
                } break;
                    
                //Case 4: (countz<remz && countn<remn && countp=remp)
                //Maximum number of +sprad reached
                case 4: {
                    
                    child[difpos.elementAt(i)] = constr1CrossSingleGene(par1, 
                            par2, i, sprad, difpos);
                    
                    //Update counters
                    if(child[difpos.elementAt(i)] == -sprad) {
                        countn++;
                    } else {
                        countz++;
                    }
                    
                } break;
                    
                //Case 5: (countz=remz && countn<remn && countp=remp)
                //Maximum number of 0s and +sprad reached
                case 5: {
                    
                    //Copy -sprad in the child
                    child[difpos.elementAt(i)] = -sprad;
                    
                } break;
                    
                //Case 1: (countz<remz && countn=remn && countp=remp)
                //Maximum number of -sprad and +sprad reached
                case 6: {
                    
                    //Copy 0 in the child
                    child[difpos.elementAt(i)] = 0;
                    
                } break;
                    
            }
            
        }
        
        return child;
        
    }
    
}
