package softcomp.genalg;

/**
 * Class containing fitness functions implementations.
 * 
 * @author Luca Mariot
 * @version 1.0
 */

import lowlevelfunc.boolfun.*;
import java.util.Random;
import java.util.Vector;

public class FitnessFunctions {
    
    /** Find the position of the maximum fitness value
     * 
     * @param fitnesses
     * @return 
     */
    public static int findMaxFitIndivPos(double[] fitnesses) {
        
        int maxfitpos = 0;
        
        for(int i=1; i<fitnesses.length; i++) {
            
            if(fitnesses[i] >= fitnesses[maxfitpos]) {
                maxfitpos = i;
            }
            
        }
        
        return maxfitpos;
        
    }
    
    /** Find the position of the minimum fitness value
     * 
     * @param fitnesses
     * @return 
     */
    public static int findMinFitIndivPos(double[] fitnesses) {
        
        int minfitpos = 0;
        
        for(int i=1; i<fitnesses.length; i++) {
            
            if(fitnesses[i] < fitnesses[minfitpos]) {
                minfitpos = i;
            }
            
        }
        
        return minfitpos;
        
    }
    
    /**
     * Compute the fitness of a pseudo-boolean function as the distance from
     * the nearest boolean function.
     * 
     * @param pseudobf truth table of a pseudo-boolean function
     * @return the distance from the nearest boolean function
     */
    public static double calcBoolDev(int[] pseudobf) {
        
        int deviation = 0;
        
        for(int i=0; i<pseudobf.length; i++) {
            
            //Compute the value of the nearest bit for the current position
            int b = 0;
            
            if(pseudobf[i] > 0) {
                
                b = 1;                          //if the value of the current 
                                                //position is higher than 0,
                                                //then collapse it to 1
            } else {
                
                if(pseudobf[i] < 0) {
                    b = -1;                     //if the value of the current 
                                                //position is lower than 0,
                                                //then collapse it to -1
                } else {
                    
                    //If the value of the current position equals 0,
                    //select a random value in {-1,+1}
                    if(Math.random()<=0.5) {
                        b = 1;
                    } else {
                        b = -1;
                    }
                }
            }
            
            //Update the fitness function
            int singledev = pseudobf[i] - b;
            deviation += (int)Math.pow(singledev,2);
            
        }
        
        return deviation;
        
    }
    
    /**
     * Evaluate the fitness of a population of restricted Walsh spectra.
     * 
     * @param pop       population of restricted Walsh spectra
     * @param speclen   length of the complete spectrum
     * @return a FitnessStatistics instance containing the fitness statistics
     *         for the current population
     */
    public static FitnessStatistics evalFitnessCombPop(WalshPermutation[] pop,
            int speclen, int[] rstspecpos, int nvar) {
        
        double maxfit = 0;
        double minfit = 0;
        double sum = 0;
        double avgfit = 0;
        double stdfit = 0;
        double varsum = 0;
        double variance = 0;
        
        int maxfitpos = 0;
        int minfitpos = 0;
        
        for(int i=0; i<pop.length; i++) {
            
            double curdev = 0;
            int[][] curfunc = CombGATools.decodeSpectrumComb(speclen, rstspecpos,
                        pop[i].getRstspec());
                
            curdev = calcBoolDev(curfunc[0]);
            
            //Update max fitness
            if(curdev > maxfit) {
                maxfit = curdev;
                maxfitpos = i;
            }
            
            //Update min fitness
            if(i==0) {
                minfit = curdev;
                minfitpos = 0;
            }
            
            if(curdev < minfit) {
                minfit = curdev;
                minfitpos = i;
            }
            
            //Update fitness accumulator
            sum += curdev;
            
            //Set the fitness of the current individual
            pop[i].setFitness(curdev);
            
        }
        
        //Update average fitness and standard deviation
        avgfit = sum / pop.length;
        
        for(int i=0; i<pop.length; i++) {
            
            double diff = avgfit - pop[i].getFitness(); 
            varsum += Math.pow(diff,2);            
            
        }
        
        variance = varsum / pop.length;
        stdfit = Math.sqrt(variance);
        
        //Build the FitnessStatistics object to be returned
        FitnessStatistics stats = new FitnessStatistics(maxfit, minfit, avgfit,
                stdfit, maxfitpos, minfitpos);
        
        return stats;
        
    }
    
}
