package softcomp.genalg;

/**
 * Main class for the permutation Genetic Algorithm
 * 
 * @author Luca Mariot
 */
import java.util.Vector;
import java.util.Random;
import java.math.BigInteger;
import lowlevelfunc.boolfun.*;
import stat.StatTools;

public class GATest {
    
    public static void main(String[] args) {
        
        //Check number of command line arguments
        if(args.length != 8) {
            
            System.err.println("\nUsage: java softcomp.genalg.TestGAComb nvar "
                    + "popsize maxgen tournsize pcross pmut sel_meth runs");
            System.exit(1);
            
        }
        
        //Read the GA parameters
        int nvar = Integer.parseInt(args[0]);
        int popsize = Integer.parseInt(args[1]);
        int maxgen = Integer.parseInt(args[2]);
        int tsize = Integer.parseInt(args[3]);
        double pcross = Double.parseDouble(args[4]);
        double pmut = Double.parseDouble(args[5]);
        int selmeth = Integer.parseInt(args[6]);
        int runs = Integer.parseInt(args[7]);
        
        //Instantiate the random number generator
        Random genrand = new Random();
        
        //Step 1: build plateaued function profile
        int[] platprof = CombGATools.buildPlatProfile(nvar);
        int m = platprof[1];
        int d = platprof[2];
        int nl = platprof[3];
        int r = platprof[4];
        int sprad = platprof[5];
        
        //Print Information
        System.out.println("\nFunction Profile: ("+nvar+","+m+","+d+","+nl+")");
        System.out.println("\nGA Information:");
        System.out.println("Population size: "+popsize);
        System.out.println("Maximum number of generations: "+maxgen);
        System.out.println("Number of independent runs: "+runs);
        System.out.println("Selection method: "+selmeth);
        System.out.println("Crossover probability: "+pcross);
        System.out.println("Mutation probability: "+pmut);
        
        //Step 2: find the permutable positions
        int resz = CombGATools.countResZeros(nvar, m);
        int speclen = (int)Math.pow(2,nvar);
        int rstspeclen = speclen - resz;
        int[] rstspecpos = CombGATools.findPermutablePos(nvar, m, rstspeclen);
        
        //Step 3: compute the proportions of zeros, -Wmax and +Wmax
        int[] whcoeffsprop = CombGATools.countWHCoeffsProp(nvar, sprad, speclen, rstspeclen);
        int propremz = whcoeffsprop[0];
        int propnzneg = whcoeffsprop[1];
        int propnzpos = whcoeffsprop[2];
        
        int numopt = 0;
        double[] stgen = new double[runs];
        
        //Cycle on Runs
        double[] bestfits = new double[runs];
        double[] fiteval = new double[runs];
        
        for(int i =0; i<runs; i++) {
            //Step 4: create a random initial population of restricted spectra
            //and evaluate their fitness
            WalshPermutation[] pop = CombGATools.createWalshPopComb(popsize,
                    propremz, propnzneg, propnzpos, sprad, rstspeclen, genrand);
            Vector<FitnessStatistics> stats = new Vector<FitnessStatistics>();
            FitnessStatistics initstat = FitnessFunctions.evalFitnessCombPop(
                    pop, speclen, rstspecpos, nvar);
            stats.add(initstat);

            //Step 4a: Update best individual
            int[] bestperm = new int[rstspeclen];
            System.arraycopy(pop[initstat.getMinfitpos()].getRstspec(), 0,
                    bestperm, 0, bestperm.length);
            WalshPermutation bestind = new WalshPermutation(bestperm);
            bestind.setFitness(pop[initstat.getMinfitpos()].getFitness());

            //Step 5: Main cycle. Evolve the population until a boolean function
            //is obtained or stopgen generations without improvements have been done
            int gen = 0;

            while((bestind.getFitness() > 0) && (gen < maxgen)) {

                //Step 5a: create the new generation
                pop = CombGATools.createNextGenComb(pop, genrand, pcross, pmut,
                        tsize, selmeth, propremz, propnzneg, propnzpos, sprad);

                //Step 5b: evaluate the fitness of the new generation
                FitnessStatistics curstat = FitnessFunctions.evalFitnessCombPop(
                    pop, speclen, rstspecpos, nvar);
                stats.add(curstat);

                //Step 5c: update best individual
                if(curstat.getMinfitval() < bestind.getFitness()) {

                    //System.out.println("Best individual changed, fitness: "+curstat.getMinfitval()+" generations: "+gen);
                    int[] bestcurgen = new int[rstspeclen];
                    System.arraycopy(pop[curstat.getMinfitpos()].getRstspec(), 0,
                        bestcurgen, 0, bestcurgen.length);
                    bestind.setRstspec(bestcurgen);
                    bestind.setFitness(curstat.getMinfitval());

                }

                //Step 5d: apply elitism (replace random individual with a
                //random one)
                int[] bestvec = new int[rstspeclen];
                System.arraycopy(bestind.getRstspec(), 0, bestvec, 0, bestvec.length);
                WalshPermutation toReplace = new WalshPermutation(bestvec);
                toReplace.setFitness(bestind.getFitness());
                int pos = genrand.nextInt(popsize);
                pop[pos] = toReplace;

                //Step 5e: update counter
                gen++;


            }

            //Step 6: print best individual information
            int[][] func = CombGATools.decodeSpectrumComb(speclen, rstspecpos,
                    bestind.getRstspec());
            System.out.print("\nRun "+(i+1)+" ; Generations performed: "+gen+" ; Best Individual "
                    + "Fitness: "+bestind.getFitness()+" ; ");
            
            //if the function is boolean, print its Wolfram number
            if(bestind.getFitness() == 0.0) {

                BigInteger funcnum = BinTools.bin2DecBig(BinTools.pol2Bin(func[0]));
                System.out.print(" ; Function number: "+funcnum);
                numopt++;

            }

            //otherwise, print the Wolfram number of the nearest boolean function
            else {

                boolean[] nearestbf = CombGATools.collapsePseudoBoolFunc(func[0]);
                BigInteger funcnum = BinTools.bin2DecBig(nearestbf);
                System.out.print (" ; Nearest Function number: "+funcnum);

            }
            
            //Update best individual statistics
            bestfits[i] = bestind.getFitness();
            
            //Update generations statistics
            stgen[i] = gen;
            
            //Update fitness evaluations statistics
            fiteval[i] = gen * popsize;
            
        }
        
        //Compute final statistics
        double meanbest = StatTools.calcMean(bestfits);
        double minbest = bestfits[FitnessFunctions.findMinFitIndivPos(bestfits)];
        double maxbest = bestfits[FitnessFunctions.findMaxFitIndivPos(bestfits)];
        double varbest = StatTools.calcVariance(bestfits, meanbest);
        double stdbest = StatTools.calcStdDev(varbest);
        
        double meangen = StatTools.calcMean(stgen);
        double vargen = StatTools.calcVariance(stgen, meangen);
        double stdgen = StatTools.calcStdDev(vargen);
        
        double meanfv = StatTools.calcMean(fiteval);
        double varfv = StatTools.calcVariance(fiteval, meanfv);
        double stdfv = StatTools.calcStdDev(varfv);
        
        System.out.println("\n\nBest Individual Statistics: ");
        System.out.println("Mean fitness = "+meanbest);
        System.out.println("Min fitness = "+minbest);
        System.out.println("Max fitness = "+maxbest);
        System.out.println("Fitness Variance = "+varbest);
        System.out.println("Fitness Standard Deviation = "+stdbest);
        System.out.println("Number of optimal solution = "+numopt);
        
        System.out.println("\n Generations Statistics: ");
        System.out.println("Generations Mean = "+meangen);
        System.out.println("Generations Variance = "+vargen);
        System.out.println("Generations Standard Deviation = "+stdgen);
        
        System.out.println("\n Fitness Evaluations Statistics: ");
        System.out.println("Fitness Evaluations Mean = "+meanfv);
        System.out.println("Fitness Evaluations Variance = "+varfv);
        System.out.println("Fitness Evaluations Standard Deviation = "+stdfv);
        
    }
    
}
