package softcomp.genalg;

/**
 * Class containing various methods for permutations of Walsh Spectra.
 * 
 * @author Luca Mariot
 */

import lowlevelfunc.boolfun.*;
import java.util.Vector;
import java.util.Random;

public class PermTools {
    
    /**
     * Build the profile (n,m,d,nl,r,sprad) of a plateaued boolean function starting
     * from the number of input variables n.
     * 
     * @param nvar number of input variables
     * @return an array of 6 int components specifiying number of variables n,
     *         resiliency order m, algebraic degree d, nonlinearity nl, plateuau
     *         parameter r and spectral radius sprad
     */
    public static int[] buildPlatProfile(int nvar) {
        
        int[] profile = new int[6];
        profile[0] = nvar;      //First component: number of variables
   
        int r=0;    //parameter for plateaued functions profile
        
        //If nvar is even, then set r to nvar/2 + 1 (otherwise we get bent
        //functions). Else, set r to (nvar+1)/2
        if((nvar % 2) == 0) {
            r = (nvar/2) + 1;  //  
        } else {                         //--> 5th component: plateau parameter
            r = (nvar+1) / 2; //
        }
        int sprad = (int)Math.pow(2,r);
        int m = r - 2;
        int d = nvar - m - 1;
        int nl = (int)Math.pow(2,nvar-1) - sprad/2;
        
        profile[1] =  m;    //2nd component: resiliency order m
        profile[2] = d;  //3rd component: algebraic degree
        profile[3] = nl; //4th component: nonlinearity
        profile[4] = r;  //5th component: parameter r
        profile[5] = sprad; //6th component: spectral radius sprad
        
        return profile;
        
    }
    
    /**
     * *** USELESS, JUST USE speclen - findPermutablePos().capacity() ***
     * 
     * Count the number of positions set to zero in the Walsh spectrum to satisfy
     * resiliency order m.
     * 
     * @param nvar  number of input variables of the boolean function
     * @param m     resiliency order of the boolean function
     * @return the number of positions in the Walsh spectrum to be set to zero
     *         in order to meet res(m)
     */
    public static int countResZeros(int nvar, int m) {
        
        int reszeros = 1;
        for(int i=1; i<=m; i++) {
            
            reszeros += CombTools.binCoeff(nvar, i);
            
        }
        
        return reszeros;
        
    }
    
    /**
     * Return a vector containing the position in the permutable positions in
     * the Walsh spectrum. These positions are the ones having Hamming weight
     * greater than m, where m is the resiliency order.
     * 
     * @param nvar  number of input variables of the boolean function
     * @param m     resiliency order of the boolean function
     * @return a Vector of Integer containing the ordered sequence of permutable
     *         positions in the Walsh spectrum
     */
    public static Vector<Integer> findPermutablePos(int nvar, int m) {
        
        Vector<Integer> nzpos = new Vector<Integer>();
        
        //Walsh spectrum length
        int speclen = (int)Math.pow(2,nvar);
        
        for(int i=0; i<speclen; i++) {
            
            //Convert the current index in binary
            boolean[] binind = BinTools.dec2BinMod(i, nvar);
            
            //If the hamming weight of the index is higher than the resiliency
            //order, add it to the vector
            if(BinTools.hwt(binind) > m) {
                nzpos.add(i);
            }
            
        }
        
        nzpos.trimToSize();
        return nzpos;
        
    }
    
    /** Count the number of the remaining positions to be set respectively to
     * zero, -Wmax and +Wmax in the Walsh spectrum
     * 
     * @param nvar          number of input variables of the boolean function
     * @param sprad         spectral radius (Wmax) of the Walsh spectrum
     * @param speclen       Walsh spectrum length
     * @param rstspeclen    number of permutable positions in the spectrum
     * @return an int array specifying the number of positions to be set
     *         respectively to zero, -Wmax and +Wmax in the spectrum
     */
    public static int[] countWHCoeffsProp(int nvar, int sprad, int speclen,
            int rstspeclen) {
        
        int parsevalsum = (int)Math.pow(2.0, (2*nvar));     //sum yielded by Parseval's relation
        //System.out.println("Parseval = "+parsevalsum);
        int propnz = parsevalsum / (sprad*sprad);   //number of positions equal to +-Wmax
        //System.out.println("Nonzero positions = "+propnz);
        int propremz = rstspeclen - propnz; //number of zero positions for the remaining zeros
        int propnzneg = (propnz - (speclen/sprad)) / 2; //number of positions equal to -Wmax
        int propnzpos = propnzneg + (speclen/sprad);    //number of positions equal to +Wmax
        
        int[] toRet = {propremz, propnzneg, propnzpos};
        
        return toRet;
        
    }
    
    /**
     * Build the ordered spectrum.
     * 
     * @param propremz
     * @param propnzneg
     * @param propnzpos
     * @param sprad
     * @return 
     */
    public static int[] buildOrdSpec(int propremz, int propnzneg,
            int propnzpos, int sprad) {
        
        //Step 1: create the ordered spectrum vector
        Vector<Integer> ordspec = new Vector<Integer>();
        
        //Fill the remaining zeros
        for(int i=0; i<propremz; i++) {
            
            ordspec.add(0);
            
        }
        
        ordspec.trimToSize();
        
        //Fill the positions equal to -Wmax
        for(int i=0; i<propnzneg; i++) {
            
            ordspec.add(-sprad);
            
        }
        
        ordspec.trimToSize();
        
        //Fill the positions equal to +Wmax
        for(int i=0; i<propnzpos; i++) {
            
            ordspec.add(sprad);
            
        }
        
        ordspec.trimToSize();
        
        int[] ordarr = new int[ordspec.capacity()];
        
        for(int i=0; i<ordarr.length; i++) {
            
            ordarr[i] = ordspec.elementAt(i);
            
        }
        
        return ordarr;
        
    }
    
    /**
     * Create an individual for the permutation GA as a random permutation of
     * the restricted Walsh spectrum.
     * 
     * @param propremz  proportions of zeros in the restricted spectrum
     * @param propnzneg proportions of -Wmax in the restricted spectrum
     * @param propnzpos proportions of +Wmax in the restricted spectrum
     * @param sprad     spectral radius (Wmax) of the Walsh spectrum
     * @param genrand   random number generator
     * @return an int array specifying a random permutation of the restricted
     *         spectrum of a pseudoboolean plateaued function
     */
    public static int[] createRandomPermIndiv(int rstspeclen, Random genrand) {
        
        //Step 1: create the ordered position vector
        Vector<Integer> ordpos = new Vector<Integer>();
        
        for(int i=0; i<rstspeclen; i++) {
            
            ordpos.add(i);
            
        }
        
        ordpos.trimToSize();
        
        //Step 2: randomly shuffle the ordered vector to get the random
        //restricted spectrum
        int[] rstspec = new int[rstspeclen];
        
        for(int i=0; i<rstspeclen; i++) {
            
            ordpos.trimToSize();
            
            //Randomly draw a position from the ordered vector
            int randpos = genrand.nextInt(ordpos.capacity());
            
            //Copy the value in the random position of the ordered vector
            //in the current index of the restricted spectrum
            rstspec[i] = ordpos.elementAt(randpos);
            
            //Delete the value from the ordered vector
            ordpos.removeElementAt(randpos);
            
        }
        
        return rstspec;
        
    }
    
    /**
     * Create a random population of restricted Walsh spectra permutations.
     * 
     * @param popsize   Size of the population
     * @param propremz  proportions of zeros in the restricted spectrum
     * @param propnzneg proportions of -Wmax in the restricted spectrum
     * @param propnzpos proportions of +Wmax in the restricted spectrum
     * @param sprad     spectral radius (Wmax) of the Walsh spectrum
     * @param genrand   random number generator
     * @return an array of WalshPermutation objects representing the population
     */
    public static WalshPermutation[] createWalshPop(int popsize, int propremz,
            int propnzneg, int propnzpos, int sprad, int rstspeclen, Random genrand) {
        
        WalshPermutation[] population = new WalshPermutation[popsize];
        
        for(int i=0; i<popsize; i++) {
            
            int[] rstspec = PermTools.createRandomPermIndiv(rstspeclen, genrand);
            population[i] = new WalshPermutation(rstspec);
            
        }
        
        return population;
        
    }
    
    /**
     * Build the pseudo-boolean function starting from its restricted spectrum,
     * using the Inverse Walsh Transform
     * @param speclen   length of the spectrum
     * @param ordspec   ordered spectrum
     * @param nzpos     Vector of permutable positions in the spectrum
     * @param rstspec   restricted spectrum
     * @return a matrix containing the truth table of the pseudo-boolean
     *         function and its Walsh spectrum
     */
    public static int[][] decodeSpectrum(int speclen, int[] ordspec,
            Vector<Integer> nzpos, int[] rstspec) {
        
        //Step 1: create the complete spectrum
        int[] func = new int[speclen];
        int[] spectrum = new int[speclen];
        
        for(int i=0; i<nzpos.capacity(); i++) {
            
            func[nzpos.elementAt(i)] = ordspec[rstspec[i]];
            spectrum[nzpos.elementAt(i)] = ordspec[rstspec[i]];
            
        }
        
        //Step 2: apply the IWT and return the function
        BoolTransf.calcInvFWT(func, 0, speclen);
        
        int[][] toRet = {func,spectrum};
        
        return toRet;
        
    }
    
    /** 
     * Collapse a pseudo-boolean function to the nearest boolean function
     * 
     * @param pseudobf  truth table of a pseudo-boolean function
     * @return the truth table of the nearest boolean function
     */
    public static boolean[] collapsePseudoBoolFunc(int[] pseudobf) {
        
        boolean[] nearestbf = new boolean[pseudobf.length];
        
        for(int i=0; i<pseudobf.length; i++) {
            
            //Compute the value of the nearest bit for the current position
            boolean b = false;
            
            if(pseudobf[i] > 0) {
                
                b = false;                      //if the value of the current 
                                                //position is higher than 0,
                                                //then collapse it to +1 (0)
            } else {
                
                if(pseudobf[i] < 0) {
                    b = true;                   //if the value of the current 
                                                //position is lower than 0,
                                                //then collapse it to -1
                } else {
                    
                    //If the value of the current position equals 0,
                    //select a random value in {-1,+1}
                    if(Math.random()<=0.5) {
                        b = false;
                    } else {
                        b = true;
                    }
                }
            }
            
            nearestbf[i] = b;
            
        }
        
        return nearestbf;
        
    }
    
}
