package softcomp.genalg;

/**
 * Class implementing the Simulated Annealing algorithm by Clark et al.
 * 
 * @author Luca Mariot
 */

import java.math.BigInteger;
import lowlevelfunc.boolfun.*;
import java.util.Vector;
import java.util.Random;
import stat.StatTools;

public class SATest {
    
    public static void main(String[] args) {
        
        //Check number of command line arguments
        if(args.length != 7) {
            
            System.err.println("\nUsage: java softcomp.genalg.SATest nvar "
                    + "moveIL maxIL alpha inittemp runs");
            System.exit(1);
            
        }
        
        //Read the SA parameters
        int nvar = Integer.parseInt(args[0]);
        int moveil = Integer.parseInt(args[1]);
        int maxil = Integer.parseInt(args[2]);
        double alpha = Double.parseDouble(args[3]);
        double temp = Double.parseDouble(args[4]);
        int runs = Integer.parseInt(args[5]);
        
        //Instantiate the random number generator
        Random genrand = new Random();
        
        //Step 1: build plateaued function profile
        int[] platprof = PermTools.buildPlatProfile(nvar);
        int m = platprof[1];
        int d = platprof[2];
        int nl = platprof[3];
        int r = platprof[4];
        int sprad = platprof[5];
        
        //Print Information
        System.out.println("\nFunction Profile: ("+nvar+","+m+","+d+","+nl+")");
        System.out.println("\nSA Information:");
        System.out.println("Moves in inner loop: "+moveil);
        System.out.println("Number of inner loops: "+maxil);
        System.out.println("Initial temperature: "+temp);
        System.out.println("Alpha: "+alpha);
        
        int numopt = 0;
        
        //Step 2: find the permutable positions
        Vector<Integer> nzpos = PermTools.findPermutablePos(nvar, m);
        
        //Step 3: compute the proportions of zeros, -Wmax and +Wmax
        int speclen = (int)Math.pow(2,nvar);
        int rstspeclen = nzpos.capacity();
        int[] whcoeffsprop = PermTools.countWHCoeffsProp(nvar, sprad, speclen, rstspeclen);
        int propremz = whcoeffsprop[0];
        int propnzneg = whcoeffsprop[1];
        int propnzpos = whcoeffsprop[2];
        
        //Create ordered spectrum
        int[] ordspec = PermTools.buildOrdSpec(propremz, propnzneg, propnzpos, sprad);
        
        System.out.println("");
        
        double[] bestfits = new double[runs];
        double[] stil = new double[runs];
        double[] fiteval = new double[runs];
        
        for(int l = 0; l < runs; l++) {
            //Create the initial solution and evaluate its fitness
            WalshPermutation sol = new WalshPermutation(
                    PermTools.createRandomPermIndiv(rstspeclen, genrand));
            sol.setFitness(FitnessFunctions.calcBoolDev(PermTools.decodeSpectrum(
                    speclen, ordspec, nzpos, sol.getRstspec())[0]));

            //Counter for number of inner loops and number of nonimproving inner loops
            int numil = 0;
            int nonimp = 0;

            //Main SA cycle
            while((sol.getFitness() > 0) /*&& (nonimp < stopnonimp)*/ && (numil < maxil)) {

                //Save the fitness of the current solution
                double curfit = sol.getFitness();

                //Perform one inner loop
                for(int i = 0; i < moveil; i++) {

                    //Select a candidate from the neighborhood of the current
                    //solution by swapping a random pair of values in the spectrum
                    int[] cand = new int[rstspeclen];
                    System.arraycopy(sol.getRstspec(), 0, cand, 0, cand.length);
                    int j=0;
                    int k=0;
                    while(j==k) {
                        j = genrand.nextInt(cand.length);
                        k = genrand.nextInt(cand.length);
                    }
                    int tmp = cand[j];
                    cand[j] = cand[k];
                    cand[k] = tmp;

                    //Evaluate the fitness of the new solution
                    double fitcand = FitnessFunctions.calcBoolDev(
                            PermTools.decodeSpectrum(speclen, ordspec, nzpos,
                                    cand)[0]);

                    //If the candidate solution is better than the current one
                    //substitute it
                    if(fitcand < sol.getFitness()) {

                        WalshPermutation newsol = new WalshPermutation(cand);
                        newsol.setFitness(fitcand);
                        sol = newsol; 

                    } else {

                        //otherwise, substitute the current solution with probability
                        //exp(-delta/T), where delta is the difference of the fitness
                        //values and T is the temperature

                        double unif = genrand.nextDouble();
                        double fitdif = fitcand - sol.getFitness();
                        double probacc = Math.exp(-(fitdif/temp));
                        if(unif < probacc) {

                            WalshPermutation newsol = new WalshPermutation(cand);
                            newsol.setFitness(fitcand);
                            sol = newsol; 

                        }

                    }

                }

                //Update temperature and counters
                temp *= alpha;
                numil++;

            }

            //Print the best solution information
            int[][] func = PermTools.decodeSpectrum(speclen, ordspec, nzpos, sol.getRstspec());
            System.out.print("Run "+l+": Inner loops performed: "+numil+"; Best Solution fitness: "
                    +sol.getFitness()+"; ");
            
            //if the function is boolean, print its Wolfram number
            if(sol.getFitness() == 0.0) {

                BigInteger funcnum = BinTools.bin2DecBig(BinTools.pol2Bin(func[0]));
                System.out.println("Function number: "+funcnum);
                numopt++;

            }

            //otherwise, print the Wolfram number of the nearest boolean function
            else {

                boolean[] nearestbf = PermTools.collapsePseudoBoolFunc(func[0]);
                BigInteger funcnum = BinTools.bin2DecBig(nearestbf);
                System.out.println("Nearest Function number: "+funcnum);

            }
            
            //Update best statistics
            bestfits[l] = sol.getFitness();
            
            //Update loops statistics
            stil[l] = numil;
            
            //Update fitness evaluations statistics
            fiteval[l] = (numil * moveil);
        
        }
        
        //Compute final statistics
        double meanbest = StatTools.calcMean(bestfits);
        double minbest = bestfits[FitnessFunctions.findMinFitIndivPos(bestfits)];
        double maxbest = bestfits[FitnessFunctions.findMaxFitIndivPos(bestfits)];
        double varbest = StatTools.calcVariance(bestfits, meanbest);
        double stdbest = StatTools.calcStdDev(varbest);
        
        double meanil = StatTools.calcMean(stil);
        double varil = StatTools.calcVariance(stil, meanil);
        double stdil = StatTools.calcStdDev(varil);
        
        double meanfv = StatTools.calcMean(fiteval);
        double varfv = StatTools.calcVariance(fiteval, meanfv);
        double stdfv = StatTools.calcStdDev(varfv);
        
        System.out.println("\n\nBest Individual Statistics: ");
        System.out.println("Mean fitness = "+meanbest);
        System.out.println("Min fitness = "+minbest);
        System.out.println("Max fitness = "+maxbest);
        System.out.println("Fitness Variance = "+varbest);
        System.out.println("Fitness Standard Deviation = "+stdbest);
        System.out.println("Number of optimal solution = "+numopt);
        
        System.out.println("\n Inner loops Statistics: ");
        System.out.println("Inner Loops Mean = "+meanil);
        System.out.println("Inner Loops Variance = "+varil);
        System.out.println("Inner Loops Standard Deviation = "+stdil);
        
        System.out.println("\n Fitness Evaluations Statistics: ");
        System.out.println("Fitness Evaluations Mean = "+meanfv);
        System.out.println("Fitness Evaluations Variance = "+varfv);
        System.out.println("Fitness Evaluations Standard Deviation = "+stdfv);

        
    }
    
}
